import os
import pygame

from .display import EntityDisplayer
from ..constants import Action


class PlayerDisplayer(EntityDisplayer):
    def __init__(self, entity):
        super().__init__(entity)

        sprite_sheet = pygame.image.load(
            os.path.join("assets", "labyrinth", "sprites", "player.png")
        )
        frame_height = sprite_sheet.get_height() // 2
        self.player_frames = [
            sprite_sheet.subsurface(
                pygame.Rect(0, 0, sprite_sheet.get_width(), frame_height)
            ),
            sprite_sheet.subsurface(
                pygame.Rect(0, frame_height, sprite_sheet.get_width(), frame_height)
            ),
        ]

        self.idle_sprite = self.player_frames[0]
        self.moving_sprite = self.player_frames[1]

    def get_sprite(self) -> pygame.Surface:
        sprite = self.moving_sprite if self.entity.moving else self.idle_sprite

        if self.entity.face_orientation == Action.RIGHT:
            sprite = pygame.transform.flip(sprite, True, False)

        return sprite



from typing import Tuple

from ..constants import Action
from ..display.player import PlayerDisplayer


class Player:
    def __init__(self, start_position: Tuple[int, int] = None):
        self._position = list(start_position) if start_position else None
        self._rendered_position = list(start_position) if start_position else None

        self.movement_speed = 0.25  # Adjust for desired speed
        self.moving = False
        self.face_orientation = Action.LEFT  # default sprite looks left

        self.displayer = PlayerDisplayer(self)

    @property
    def position(self):
        return tuple(self._position)

    @position.setter
    def position(self, value):
        if isinstance(value, tuple):
            self._position = list(value)
        else:
            self._position = value

    @property
    def rendered_position(self):
        return tuple(self._rendered_position)

    @rendered_position.setter
    def rendered_position(self, value):
        if isinstance(value, tuple):
            self._rendered_position = list(value)
        else:
            self._rendered_position = value

    def potential_next_position(self, action):
        """Returns the potential next position based on the action, without checking its validity."""
        potential_position = list(self.position)  

        if action == Action.UP:  # Up
            potential_position[0] -= 1
        elif action == Action.RIGHT:  # Right
            potential_position[1] += 1
        elif action == Action.DOWN:  # Down
            potential_position[0] += 1
        elif action == Action.LEFT:  # Left
            potential_position[1] -= 1

        return tuple(potential_position)

    def move_render_position(self):
        """Move the rendered_position towards position."""
        new_rendered_position = list(self._rendered_position)

        for i in range(2):  # For x and y coordinates
            diff = self._position[i] - new_rendered_position[i]
            if abs(diff) > 0.01:  # Adjust for desired precision
                new_rendered_position[i] += diff * self.movement_speed

        self._rendered_position = new_rendered_position

    def _positions_are_close(self, pos1, pos2, threshold=0.015):
        return abs(pos1[0] - pos2[0]) < threshold and abs(pos1[1] - pos2[1]) < threshold



import numpy as np
import random
from collections import deque
from queue import PriorityQueue
from typing import List, Tuple, Union, Optional

from ..constants import WALL, PATH, START, TARGET, CorridorMoveStatus

from ...common.grid_functions import on_line


class CorridorBuilder:
    def __init__(self, maze):
        self.maze = maze
        self.rows = self.maze.rows
        self.cols = self.maze.cols

    ########## Prim's algorithm logic ########################
    def generate_corridor_prim(self):
        """Generate corridors using an adaptation of Prim's algorithm.

        Builds the corridor around the rooms, without touching the rooms, so that
        the access points of the rooms need to be connected to the corridors.
        """
        grid = np.full((self.rows, self.cols), WALL)
        directions = [(2, 0), (-2, 0), (0, 2), (0, -2)]

        # Initialize the starting position
        current_position = self.maze.start_position
        grid[current_position] = PATH

        walls = []

        for d in directions:
            pos_valid, move_status = self.is_valid_next_position(
                grid, current_position, d
            )
            if pos_valid or move_status in [
                CorridorMoveStatus.ROOM_BOUNDARY,
                CorridorMoveStatus.MAZE_BOUNDARY,
            ]:
                walls.append((current_position, d))

        while walls:
            # Randomly select a wall from the list
            current_position, direction = self.maze.py_random.choice(walls)

            next_pos = (
                current_position[0] + direction[0],
                current_position[1] + direction[1],
            )

            intermediary_pos = (
                current_position[0] + direction[0] // 2,
                current_position[1] + direction[1] // 2,
            )

            is_valid, status = self.is_valid_next_position(
                grid, current_position, direction
            )

            if is_valid:
                grid[intermediary_pos[0], intermediary_pos[1]] = PATH
                grid[next_pos[0], next_pos[1]] = PATH

                for d in directions:
                    pos_valid, move_status = self.is_valid_next_position(
                        grid, next_pos, d
                    )
                    if pos_valid:
                        walls.append((next_pos, d))

            if not is_valid and self.maze.corridor_grid_connect_option:
                if (
                    0 <= intermediary_pos[0] < self.rows
                    and 0 <= intermediary_pos[1] < self.cols
                    and grid[intermediary_pos[0], intermediary_pos[1]] == WALL
                    and not self.is_adjacent_to_room(intermediary_pos)
                ):
                    grid[intermediary_pos[0], intermediary_pos[1]] = PATH

            # Remove the wall from the list
            walls.remove((current_position, direction))

        return grid

    def is_out_of_bounds_maze(self, position):
        if (
            position[0] < 0
            or position[0] >= self.rows
            or position[1] < 0
            or position[1] >= self.cols
        ):
            return True
        return False

    def is_valid_next_position(self, grid, position, direction):
        new_pos = (position[0] + direction[0], position[1] + direction[1])

        # Detect if new position is out of maze bounds
        if self.is_out_of_bounds_maze(new_pos):
            return (False, CorridorMoveStatus.MAZE_BOUNDARY)

        # Detect if new position or its neighbors are a path (either corridor or room)
        if (
            grid[new_pos[0], new_pos[1]] == PATH
            or self.maze.room_grid[new_pos[0], new_pos[1]] == PATH
        ):
            return (False, CorridorMoveStatus.INVALID)

        # Check for room padding
        if self.is_adjacent_to_room(new_pos):
            return (False, CorridorMoveStatus.ROOM_BOUNDARY)

        return (True, CorridorMoveStatus.VALID_MOVE)

    def is_adjacent_to_room(self, position):
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if (
                    0 <= position[0] + dx < self.rows
                    and 0 <= position[1] + dy < self.cols
                    and self.maze.room_grid[position[0] + dx, position[1] + dy] == PATH
                ):
                    return True
        return False
    
    def greedy_sort(self, access_points, start_position):
        # Initialize the sorted list with the start_position
        sorted_points = [start_position]
        
        # Remaining points to be sorted
        remaining_points = access_points.copy()
        
        while remaining_points:
            current_point = sorted_points[-1] # Get the last added point
            closest_point = min(remaining_points, key=lambda point: self.heuristic(current_point, point))
            
            sorted_points.append(closest_point)
            remaining_points.remove(closest_point)
        
        # Return the sorted list without the start_position
        return sorted_points[1:]

    ####### Greedy best first search logic #########
    def generate_corridor_gbfs(self):
        """Generate corridors using a slight variation of greedy best first search algorithm.
        Instead of using only the heuristic we are also considering the cost matrix.
        """

        # Initialize a grid filled with walls
        corridor_grid = np.full((self.rows, self.cols), WALL)

        # Collect all access points from the rooms
        access_points = []
        for room in self.maze.rooms:
            access_points_global = [
                (x + room.global_position[0], y + room.global_position[1])
                for x, y in room.access_points
            ]
            access_points.extend(access_points_global)

        current_endpoint = self.maze.start_position
        cost_grid = self.make_cost_grid(access_points, corridor_grid)

        # Sort the access points to get even more direct paths
        if self.maze.corridor_sort_access_points_option:
            access_points = self.greedy_sort(access_points, current_endpoint)
        else:
            self.maze.py_random.shuffle(access_points)
            
        for access_point in access_points:
            path = self.gbfs(current_endpoint, access_point, cost_grid)
            for point in path:
                corridor_grid[point[0], point[1]] = PATH
            current_endpoint = access_point
            cost_grid = self.make_cost_grid(access_points, corridor_grid)

        return corridor_grid
    
    def gbfs(self, start, goal, cost_grid):
        frontier = PriorityQueue()
        frontier.put(start, 0)

        came_from = {}
        came_from[start] = None

        while not frontier.empty():
            current = frontier.get()

            if current == goal:
                break

            for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                next_node = (current[0] + dx, current[1] + dy)

                if self.is_out_of_bounds_maze(next_node):
                    continue  # Out of maze boundary

                if next_node not in came_from:
                    priority = self.heuristic(goal, next_node, cost_grid)
                    frontier.put(next_node, priority)
                    came_from[next_node] = current

        # Reconstruct the path
        current = goal
        path = []
        while current != start:
            path.append(current)
            current = came_from[current]
        path.append(start)
        path.reverse()

        return path

    ####### A* direct corridor generation logic #########
    def generate_corridor_a_star(self):
        """Generate corridors using the A* algorithm."""

        # Initialize a grid filled with walls
        corridor_grid = np.full((self.rows, self.cols), WALL)

        # Collect all access points from the rooms
        access_points = []
        for room in self.maze.rooms:
            access_points_global = [
                (x + room.global_position[0], y + room.global_position[1])
                for x, y in room.access_points
            ]
            access_points.extend(access_points_global)

        current_endpoint = self.maze.start_position
        cost_grid = self.make_cost_grid(access_points, corridor_grid)
        
        # Sort the access points to get even more direct paths
        if self.maze.corridor_sort_access_points_option:
            access_points = self.greedy_sort(access_points, current_endpoint)
        else:
            self.maze.py_random.shuffle(access_points)

        for access_point in access_points:
            path = self.a_star_path(current_endpoint, access_point, cost_grid)
            for point in path:
                corridor_grid[point[0], point[1]] = PATH
            current_endpoint = access_point
            cost_grid = self.make_cost_grid(access_points, corridor_grid)

        return corridor_grid

    def make_cost_grid(self, access_points, corridor_grid):
        cost_grid = np.full((self.rows, self.cols), 1)

        global_room_mask = self.maze.generate_global_room_mask()
        cost_grid[global_room_mask] = 1_000_000

        for room in self.maze.rooms:
            global_pos = room.global_position
            perimeter_cells = room.get_perimeter_cells(padding=1)
            for cell in perimeter_cells:
                if self.is_next_to_access_point(cell, access_points):
                    cost_grid[global_pos[0] + cell[0], global_pos[1] + cell[1]] = 0
                else:
                    cost_grid[global_pos[0] + cell[0], global_pos[1] + cell[1]] = 100

            for access_point in room.access_points:
                cost_grid[
                    global_pos[0] + access_point[0], global_pos[1] + access_point[1]
                ] = 0

        return cost_grid

    def a_star_path(self, start, goal, cost_grid):
        frontier = PriorityQueue()
        frontier.put(start, 0)

        came_from = {}
        came_from[start] = None
        cost_so_far = {}
        cost_so_far[start] = 0

        visited = set()
        visited.add(start)

        while not frontier.empty():
            current = frontier.get()

            if current == goal:
                break

            for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                next_node = (current[0] + dx, current[1] + dy)

                if next_node in visited:
                    continue

                if self.is_out_of_bounds_maze(next_node):
                    continue

                cell_cost = cost_grid[next_node[0], next_node[1]]
                new_cost = cost_so_far[current] + cell_cost
                if next_node not in cost_so_far or new_cost < cost_so_far[next_node]:
                    cost_so_far[next_node] = new_cost
                    priority = new_cost + self.heuristic(goal, next_node)
                    frontier.put(next_node, priority)
                    came_from[next_node] = current

        # Reconstruct the path
        current = goal
        path = []
        while current != start:
            path.append(current)
            current = came_from[current]
        path.append(start)
        path.reverse()

        return path


    def is_next_to_access_point(self, position, access_points=None):
        """Checks if the given position is adjacent to any of the room's access points."""
        if access_points is None:
            access_points = []
            for room in self.maze.rooms:
                access_points_global = [
                    (x + room.global_position[0], y + room.global_position[1])
                    for x, y in room.access_points
                ]
                access_points.extend(access_points_global)

        for access_point in access_points:
            for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                if position == (access_point[0] + dx, access_point[1] + dy):
                    return True
        return False

    def heuristic(self, goal, next_node, cost_grid=None):
        """Manhattan distance heuristic."""
        if cost_grid is None:
            return abs(goal[0] - next_node[0]) + abs(goal[1] - next_node[1])
        else:
            return abs(goal[0] - next_node[0]) + abs(goal[1] - next_node[1]) + cost_grid[next_node[0], next_node[1]]

    ####### Connect access points to paths ############

    def connect_rooms_to_paths(self):
        """Connect the rooms to the corridors via the shortest path."""

        for room in self.maze.rooms:
            for access_point in room.access_points:
                global_access_point = (
                    room.global_position[0] + access_point[0],
                    room.global_position[1] + access_point[1],
                )
                connection = self.bfs_to_find_closest_path(global_access_point)

                if not connection:
                    # If BFS failed to find a connection, try growing a path
                    connection = self.grow_path_from(global_access_point)

                # If we found a connection, draw a path to it
                if connection:
                    self.plot_path_from_to(global_access_point, connection)

    def bfs_to_find_closest_path(self, access_point):
        directions = [(1, 0), (-1, 0), (0, 1), (0, -1)]

        visited = np.zeros((self.rows, self.cols), dtype=bool)
        queue = deque([access_point])

        # If the access point is already on a corridor path
        if self.maze.corridor_grid[access_point[0], access_point[1]] == PATH:
            return access_point

        while queue:
            current = queue.popleft()

            for d in directions:
                new_pos = (current[0] + d[0], current[1] + d[1])

                # Check boundaries
                if 0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols:
                    # If the new position is an existing path, we found a connection
                    if self.maze.corridor_grid[new_pos[0], new_pos[1]] == PATH:
                        return new_pos

                    # If not visited, not a room cell, not inside any room (excluding the access point), and not yet part of the corridor
                    if (
                        not visited[new_pos[0], new_pos[1]]
                        and not self.maze.is_inside_any_room(
                            new_pos, exception=access_point
                        )
                        and self.maze.corridor_grid[new_pos[0], new_pos[1]] == WALL
                    ):
                        visited[new_pos[0], new_pos[1]] = True
                        queue.append(new_pos)

        return None

    def plot_path_from_to(self, p1, p2):
        # Directly connect if the cells are adjacent and not diagonally so
        diff_x = abs(p1[0] - p2[0])
        diff_y = abs(p1[1] - p2[1])
        if diff_x <= 1 and diff_y <= 1 and not (diff_x == 1 and diff_y == 1):
            self.maze.corridor_grid[p2[0], p2[1]] = PATH
            return

        while p1 != p2:
            # First adjust horizontally
            if p1[1] != p2[1]:
                p1 = (p1[0], p1[1] + np.sign(p2[1] - p1[1]))
            # Then adjust vertically
            else:
                p1 = (p1[0] + np.sign(p2[0] - p1[0]), p1[1])

            if not self.maze.is_inside_any_room(p1, exception=None):
                self.maze.corridor_grid[p1[0], p1[1]] = PATH

    ####### Grow path logic #########
    def grow_path_from(
        self, start_pos, max_attempts=5000
    ):  # Temporarily increased max_attempts
        """Grow a path from start_pos until it reaches a corridor or max_attempts are reached."""

        current_pos = start_pos
        attempt = 0
        directions = [(0, 1), (0, -1), (1, 0), (-1, 0)]

        visited = set()
        visited.add(start_pos)

        # Get a list of all corridor positions
        corridor_positions = np.argwhere(self.maze.corridor_grid == PATH)

        while attempt < max_attempts:
            potential_directions = []

            # Get the closest corridor position to the current position
            distances = [
                np.linalg.norm(np.array(current_pos) - np.array(corridor_pos))
                for corridor_pos in corridor_positions
            ]
            closest_corridor_pos = corridor_positions[np.argmin(distances)]

            for d in directions:
                new_pos = (current_pos[0] + d[0], current_pos[1] + d[1])

                if new_pos in visited:
                    continue

                # Check boundaries
                if not (0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols):
                    continue  # Go to the next direction if this is out of bounds

                # If the new position is an existing path, validate and then return
                if self.maze.corridor_grid[new_pos[0], new_pos[1]] == PATH:
                    diff_x = new_pos[0] - current_pos[0]
                    diff_y = new_pos[1] - current_pos[1]

                    # If this is not diagonal, then validate
                    if not (abs(diff_x) == 1 and abs(diff_y) == 1):
                        if not self.is_line_segment_intersecting_room(
                            start_pos, new_pos
                        ):
                            return new_pos
                    continue  # Go to the next direction

                # If not inside any room and not yet part of the corridor
                if (
                    not self.maze.is_inside_any_room(new_pos)
                    and self.maze.corridor_grid[new_pos[0], new_pos[1]] == WALL
                ):
                    potential_directions.append(
                        (
                            new_pos,
                            self.direction_cost(current_pos, d, closest_corridor_pos),
                        )
                    )

            # Sort the potential directions by their cost
            potential_directions.sort(key=lambda x: x[1])

            if not potential_directions:
                break

            # Select the best direction
            next_pos = potential_directions[0][0]
            visited.add(next_pos)

            # Update the grid
            self.maze.corridor_grid[next_pos[0], next_pos[1]] = PATH
            current_pos = next_pos
            attempt += 1

        # If we reached here, we didn't find a corridor
        return None

    def direction_cost(self, current_pos, direction, target_pos):
        new_pos = (current_pos[0] + direction[0], current_pos[1] + direction[1])
        base_cost = abs(new_pos[0] - target_pos[0]) + abs(new_pos[1] - target_pos[1])

        # Penalize if the new position is close to a room.
        if self.maze.is_inside_any_room(new_pos):
            base_cost += 10

        return base_cost

    def is_line_segment_intersecting_room(self, p1, p2):
        """Check if the line segment p1p2 intersects any room cell."""
        for room in self.maze.rooms:
            for i in range(room.rows):
                for j in range(room.cols):
                    if room.grid[i][j] == PATH:
                        global_cell_pos = (
                            room.global_position[0] + i,
                            room.global_position[1] + j,
                        )

                        # Skip if it's the starting point
                        if global_cell_pos == p1:
                            continue

                        if on_line(p1, global_cell_pos, p2):
                            return True
        return False

    ###### Post Processing logic ######

    def post_process_maze(self):
        # Initialize the global room mask
        global_room_mask = self.maze.generate_global_room_mask()

        # Collect border cells
        border_cells = []

        for i in range(self.rows):
            for j in [0, self.cols - 1]:  # Left and right borders
                border_cells.append((i, j))

        for j in range(self.cols):
            for i in [0, self.rows - 1]:  # Top and bottom borders
                border_cells.append((i, j))

        self.maze.py_random.shuffle(border_cells)  # Shuffle border cells

        for cell in border_cells:
            self._attempt_fill_path(cell, global_room_mask)

        # Collect the perimeter cells of all rooms
        room_perimeter_cells = []

        for room in self.maze.rooms:
            perimeter_cells = room.get_perimeter_cells(padding=2)
            for cell in perimeter_cells:
                # Translate to global maze coordinates
                global_pos = (
                    cell[0] + room.global_position[0],
                    cell[1] + room.global_position[1],
                )
                room_perimeter_cells.append(global_pos)

        self.maze.py_random.shuffle(room_perimeter_cells)  # Shuffle perimeter cells

        for cell in room_perimeter_cells:
            self._attempt_fill_path(cell, global_room_mask)

    def _attempt_fill_path(self, pos, global_room_mask):
        if pos[0] < 0 or pos[0] >= self.rows or pos[1] < 0 or pos[1] >= self.cols:
            return

        adjacent_dirs = [(0, 1), (0, -1), (1, 0), (-1, 0)]
        all_directions = adjacent_dirs + [(1, 1), (-1, 1), (-1, -1), (1, -1)]

        path_count = sum(
            [
                1
                for dx, dy in adjacent_dirs
                if 0 <= pos[0] + dx < self.rows
                and 0 <= pos[1] + dy < self.cols
                and self.maze.corridor_grid[pos[0] + dx, pos[1] + dy] == PATH
            ]
        )

        inside_or_adjacent_to_room = False
        for dx, dy in all_directions:
            new_pos = (pos[0] + dx, pos[1] + dy)
            if 0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols:
                if global_room_mask[new_pos[0], new_pos[1]]:
                    inside_or_adjacent_to_room = True
                    break

        if not inside_or_adjacent_to_room and path_count == 1:
            self.maze.corridor_grid[pos[0], pos[1]] = PATH



import os
import pygame
from abc import ABC, abstractmethod

from ..constants import *

class EntityDisplayer(ABC):
    def __init__(self, entity):
        self.entity = entity

    @abstractmethod
    def get_sprite(self) -> pygame.Surface:
        """
        Return the pygame surface representing the current state of the entity.
        """
        pass

class EnvDisplay:
    STATE_COLORS = {
        WALL: (0, 0, 0),
        PATH: (255, 255, 255),
        PLAYER: (0, 255, 0),
        TARGET: (255, 0, 0),
        START: (0, 0, 255),
    }
    GRID_COLOR = (200, 200, 200)  # A light gray color for grid lines
    BORDER_COLOR = (50, 50, 50)  # A darker gray for the maze border
    BORDER_PADDING = 5

    def __init__(
        self,
        rows: int,
        cols: int,
        window_width: int = 800,
        window_height: int = 800,
        labyrinth=None,
    ):
        """
        Args:
            rows (int): The number of rows in the display.
            cols (int): The number of columns in the display.
            window_width (int): Width of the window.
            window_height (int): Height of the window.
            labyrinth (Laabyrinth): The labyrinth to display.
        """
        self.rows = rows
        self.cols = cols
        self.window_width = window_width
        self.window_height = window_height
        self.cell_width = None
        self.cell_height = None
        self.additional_padding_x = None
        self.additional_padding_y = None
        self.labyrinth = labyrinth

        self.target_sprite = pygame.image.load(
            os.path.join("assets", "labyrinth", "sprites", "flag.png")
        )

        self.player_displayer = self.labyrinth.player.displayer
        
        self._compute_sizes_and_paddings()

        self.screen = pygame.display.set_mode(
            (self.window_width, self.window_height), pygame.RESIZABLE
        )

        pygame.display.set_caption("Labyrinth")

    def _compute_sizes_and_paddings(self):
        self.cell_width = (self.window_width - 2 * EnvDisplay.BORDER_PADDING) // self.cols
        self.cell_height = (
            self.window_height - 2 * EnvDisplay.BORDER_PADDING
        ) // self.rows

        total_cell_width = round(self.cell_width * self.cols)
        total_cell_height = round(self.cell_height * self.rows)

        self.additional_padding_x = (
            self.window_width - round(total_cell_width) - 2 * EnvDisplay.BORDER_PADDING
        ) 
        self.additional_padding_y = (
            self.window_height - round(total_cell_height) - 2 * EnvDisplay.BORDER_PADDING
        ) 

    def draw_state(self):
        """Display the state of the env

        Args:
            state (2D Numpy array): The discrete state of the environment
            animate (bool, optional): Whether to animate movement of agents. Defaults to True.
        """
        # 1. Fill background with border color
        self.screen.fill(EnvDisplay.BORDER_COLOR)

        # 2. Draw maze
        self.draw_maze(self.labyrinth.maze.grid)

        # 3. Draw target
        self.draw_target_at_position(self.labyrinth.maze.target_position)

        # 4. Draw the grid
        self.draw_grid(self.labyrinth.maze.grid.shape[0], self.labyrinth.maze.grid.shape[1])

        #### Draw special features
        # 5. Draw player
        self.draw_player()

        pygame.display.flip()

    def draw_player(self):
        draw_pos = self.labyrinth.player.rendered_position
            
        x, y = self._adjust_coords_for_padding(*draw_pos)

        sprite = self.player_displayer.get_sprite()

        # Scale the sprite
        desired_width = self.cell_width * 0.8
        desired_height = self.cell_height * 0.8
        sprite = pygame.transform.scale(sprite, (desired_width, desired_height))

        # micro-adjustement
        x = x + (self.BORDER_PADDING / 2)
        y = y + (self.BORDER_PADDING / 2)
        self.screen.blit(sprite, (x, y))

    

    def draw_target_at_position(self, position):
        x, y = self._adjust_coords_for_padding(*position)
        sprite = self.get_target_sprite()
        self.screen.blit(sprite, (x, y))

    def get_target_sprite(self):
        desired_width = self.cell_width * 0.9
        desired_height = self.cell_height * 0.9
        sprite = pygame.transform.scale(
            self.target_sprite, (desired_width, desired_height)
        )
        return sprite

    def _adjust_coords_for_padding(self, row, col):
        """
        Translate a logical grid position (row, col) to pixel coordinates.
        """
        x = self.additional_padding_x + col * self.cell_width
        y = self.additional_padding_y + row * self.cell_height
        return x, y

    def draw_maze(self, state):
        for row in range(len(state)):
            for col in range(len(state[row])):
                cell_value = state[row][col]
                if cell_value == WALL:
                    self.draw_cell(row, col, WALL)
                else:
                    self.draw_cell(row, col, PATH)

    def draw_cell(self, row, col, cell_value):
        color = self.STATE_COLORS.get(cell_value)

        # Adjust for padding
        x = round(col * self.cell_width) + self.additional_padding_x
        y = round(row * self.cell_height) + self.additional_padding_y

        pygame.draw.rect(
            self.screen,
            color,
            pygame.Rect(x, y, round(self.cell_width), round(self.cell_height)),
        )

    def draw_grid(self, rows, cols):
        for col in range(1, cols):
            x = col * self.cell_width + self.additional_padding_x
            pygame.draw.line(
                self.screen,
                EnvDisplay.GRID_COLOR,
                (x, self.additional_padding_y),
                (x, (rows * self.cell_height + self.additional_padding_y)),
            )

        for row in range(1, rows):
            y = row * self.cell_height + self.additional_padding_y
            pygame.draw.line(
                self.screen,
                EnvDisplay.GRID_COLOR,
                (self.additional_padding_x, y),
                ((cols * self.cell_width + self.additional_padding_x), y),
            )

    def resize(self, new_width, new_height):
        self.window_width = new_width
        self.window_height = new_height
        self._compute_sizes_and_paddings()
        self.screen = pygame.display.set_mode(
            (self.window_width, self.window_height), pygame.RESIZABLE
        )



from typing import List, Tuple, Union, Optional

import gymnasium as gym
import numpy as np
import random
import pygame

from .maze.maze import MazeFactory
from .constants import WALL, PATH, TARGET, START, PLAYER, Action
from .display.display import EnvDisplay
from .entities.player import Player


class Labyrinth(gym.Env):
    def __init__(
        self,
        rows: int,
        cols: int,
        maze_nr_desired_rooms: Optional[int] = None,
        maze_nr_desired_rooms_range: Tuple[int, int] = (1, 8),
        maze_global_room_ratio: Optional[float] = None,
        maze_global_room_ratio_range: Tuple[float, float] = (0.1, 0.8),
        maze_corridor_algorithm: Optional[str] = "random",
        maze_corridor_grid_connect_option: Union[bool, str] = "random",
        maze_corridor_post_process_option: bool = True,
        maze_corridor_sort_access_points_option: Union[bool, str] = "random",
        room_access_points: Optional[int] = None,
        room_access_points_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        room_ratio: Optional[Union[int, float]] = None,
        room_ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        reward_schema: Optional[dict] = None,
        seed: Optional[int] = None,
    ):
        """
        Labyrinth environment for reinforcement learning.

        Arguments maze_num_rooms, maze_global_room_ratio, room_access_points
        are used to fix specific values, otherwise the values are drawn from the minimum and maximum distibutions.

        room_types is used to determine what types of rooms are to be added in the maze, if None, the random seletion
        considers all the implemented room types

        Args:
            rows (int): The number of rows in the labyrinth.
            cols (int): The number of columns in the labyrinth.
            maze_nr_desired_rooms (int, optional): The desired number of rooms in the maze. Defaults to None.
            maze_nr_desired_rooms_range (tuple, optional): The range of desired number of rooms. Defaults to (1, 8).
            maze_global_room_ratio (float, optional): The global room ratio in the maze. Defaults to None.
            maze_global_room_ratio_range (tuple, optional): The range of global room ratio. Defaults to (0.1, 0.8).
            maze_corridor_algorithm (str, optional): The algorithm to use for generating the maze. Defaults to 'random'. Can be one of: 'prim', 'astar' or 'random'. if 'random', randomly selects algorithm.
            maze_corridor_grid_connect_option (Union[bool, str], optional): Option to decide the nature of corridor paths connectivity.
                - If `True`, corridors will be grid-connected.
                - If `False`, corridors will not be grid-connected.
                - If `"random"`, the choice will be made randomly.
                Defaults to False.
            maze_corridor_post_process_option (bool, optional): Whether to post-process the maze. Defaults to True.
            maze_corridor_sort_access_points_option (Union[bool, str], str, optional): Whether to sort the access points. Defaults to "random". If "random" will randomly select between "True" and "False".
            room_access_points (int, optional): The number of access points in each room. Defaults to None.
            room_access_points_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added in the maze. Defaults to None.
            room_ratio (float, optional): The room ratio. Defaults to None.
            room_ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            reward_schema (dict, optional): A dictionary defining the reward schema for the labyrinth. Defaults to None.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """

        super().__init__()

        self.rows, self.cols = rows, cols
        self.state = np.ones((rows, cols), dtype=np.uint8) * WALL

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)

        self.py_random = random.Random(self.seed)
        self.np_random = np.random.RandomState(self.seed)

        # Define action and observation spaces
        self.action_space = gym.spaces.Discrete(4)  # Up, Right, Down, Left
        self.observation_space = gym.spaces.Box(
            low=0, high=4, shape=(rows, cols), dtype=np.uint8
        )

        if not reward_schema:
            self.reward_schema = {
                "neutral_reward": -0.01,
                "wall_collision_reward": -1,
                "target_reached_reward": 10,
            }

        #### Setup maze factory settings ####
        self.maze_nr_desired_rooms = maze_nr_desired_rooms
        self.maze_nr_desired_rooms_range = maze_nr_desired_rooms_range
        self.maze_global_room_ratio = maze_global_room_ratio
        self.maze_global_room_ratio_range = maze_global_room_ratio_range
        self.maze_corridor_algorithm = maze_corridor_algorithm
        self.maze_corridor_grid_connect_option = maze_corridor_grid_connect_option
        self.maze_corridor_post_process_option = maze_corridor_post_process_option
        self.maze_corridor_sort_access_points_option = (
            maze_corridor_sort_access_points_option
        )
        self.room_access_points = room_access_points
        self.room_access_points_range = room_access_points_range
        self.room_types = room_types
        self.room_ratio = room_ratio
        self.room_ratio_range = room_ratio_range

        self.player = Player()

        self.setup_labyrinth()

        self.env_displayer = None  # only initialize if render is needed

    def setup_labyrinth(self):
        self.make_maze_factory()
        self.maze = self.maze_factory.create_maze()
        self.player.position = self.maze.start_position
        self.player.rendered_position = self.player.position
        self.build_state_matrix()

    def make_maze_factory(self):
        maze_factory_seed = self.seed + 1
        self.maze_factory = MazeFactory(
            rows=self.rows,
            cols=self.cols,
            nr_desired_rooms=self.maze_nr_desired_rooms,
            nr_desired_rooms_range=self.maze_nr_desired_rooms_range,
            global_room_ratio=self.maze_global_room_ratio,
            global_room_ratio_range=self.maze_global_room_ratio_range,
            access_points_per_room=self.room_access_points,
            access_points_per_room_range=self.room_access_points_range,
            room_types=self.room_types,
            room_ratio=self.room_ratio,
            room_ratio_range=self.room_ratio_range,
            corridor_algorithm=self.maze_corridor_algorithm,
            corridor_grid_connect_option=self.maze_corridor_grid_connect_option,
            corridor_post_process_option=self.maze_corridor_post_process_option,
            corridor_sort_access_points_option=self.maze_corridor_sort_access_points_option,
            seed=maze_factory_seed,
        )
        return self.maze_factory

    def build_state_matrix(self):
        """Sequentially build the state matrix."""
        self.state = self.maze.grid.copy()
        self.state[self.maze.start_position] = START
        self.state[self.maze.target_position] = TARGET
        self.state[self.player.position] = PLAYER
        return self.state

    def step(self, action):
        action = Action(action)

        # Initial reward
        reward = self.reward_schema["neutral_reward"]
        done = False
        truncated = False

        # Make the agent sprite turn
        if action == Action.LEFT or action == Action.RIGHT:
            self.player.face_orientation = action

        # Check if the action is valid
        if not self.is_valid_move(self.player, action):
            reward = self.reward_schema["wall_collision_reward"]
            return self.state, reward, done, truncated, {"info": "Invalid move!"}

        # Move the agent
        self.agent_move(action)

        # Check if the agent reached the target
        if self.player.position == self.maze.target_position:
            reward = self.reward_schema["target_reached_reward"]
            done = True
            return self.state, reward, done, truncated, {"info": "Reached the target!"}

        # Flush all info to state matrix
        self.state = self.build_state_matrix()

        return self.state, reward, done, truncated, {}

    def is_valid_move(self, player, action):
        potential_position = player.potential_next_position(action)
        is_inside_bounds = (0 <= potential_position[0] < self.rows) and (
            0 <= potential_position[1] < self.cols
        )
        if not is_inside_bounds:
            return False
        if self.maze.grid[potential_position[0], potential_position[1]] == WALL:
            return False
        return True

    def agent_move(self, action):
        new_position = self.player.potential_next_position(action)
        self.player.position = new_position

    def reset(self, seed=None):
        """Reset and regenerate another labyrinth. If the same seed as the one at the initialization is provided,
        then the same labyrinth should be regenerated.

        Args:
            seed (int, optional): External seed if a user wants to provide one. Defaults to None.

        """
        if seed:
            self.seed = seed
        else:
            self.seed += 1

        self.setup_labyrinth()

    def render(
        self,
        sleep_time: int = 100,
        window_size: Tuple[int, int] = (800, 800),
        animate: bool = True,
        process_arrow_keys: bool = False,
    ):
        """
        Renders the environment and handles user input events.

        Args:
            sleep_time (int, optional): The sleep time in milliseconds between each frame of the animation. Defaults to 100.
            window_size (Tuple[int, int], optional): The size of the window in pixels. Defaults to (800, 800).
            animate (bool, optional): Whether to animate the rendering. Defaults to True.
            process_arrow_keys (bool, optional): Whether to process directional keys for user input. Defaults to False.

        Returns:
            Tuple[bool, Action]: A tuple containing a boolean value indicating if the quit event occurred and the action taken by the user.
        """
        if self.env_displayer is None:
            # Initialize only if render is needed
            self.env_displayer = EnvDisplay(
                self.rows,
                self.cols,
                window_width=window_size[0],
                window_height=window_size[1],
                labyrinth=self,
            )

        quit_event = False
        action = None

        if not animate:
            self.player.rendered_position = self.player.position

        self.env_displayer.draw_state()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                quit_event = True
                return quit_event, action

            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    pygame.quit()
                    quit_event = True
                    return quit_event, action

                if process_arrow_keys:
                    if event.key == pygame.K_UP:
                        action = Action.UP
                    elif event.key == pygame.K_RIGHT:
                        action = Action.RIGHT
                    elif event.key == pygame.K_DOWN:
                        action = Action.DOWN
                    elif event.key == pygame.K_LEFT:
                        action = Action.LEFT

            if event.type == pygame.VIDEORESIZE:
                self.env_displayer.resize(event.w, event.h)
                self.env_displayer.draw_state()

        if animate:
            # This animation effect makes the display slightly unresponsive as
            # it blocks the execution here
            while not self.player._positions_are_close(
                self.player.rendered_position, self.player.position
            ):
                self.player.moving = True
                self.player.move_render_position()
                self.env_displayer.draw_state()
                pygame.time.wait(
                    int(sleep_time / 10)
                )  # Faster refresh for smoother animation

            self.player.moving = False

        return quit_event, action

    def human_play(self, print_info=False, window_size=(800, 800), animate=True):
        """Continously display environment and allow user to play.
        Exit by closing the window or pressing ESC.
        """
        self.render(window_size=window_size, animate=animate, process_arrow_keys=True)

        done = False
        first_info_printed = True
        while True:
            reward, done, info = None, None, None

            quit_event, action = self.render(
                window_size=window_size, animate=animate, process_arrow_keys=True
            )

            if quit_event:
                break

            if action is not None:
                _, reward, done, _, info = self.step(action)

                if print_info:
                    print(f"Reward: {reward}, Done: {done}, Info: {info}")

            if print_info and first_info_printed:
                init_message = (
                    f"Initialized environment with seed: {self.seed}, "
                    f"rows: {self.rows}, cols: {self.cols}, "
                    f"maze_nr_desired_rooms: {self.maze.nr_desired_rooms}, "
                    f"maze_nr_placed_rooms: {self.maze.nr_placed_rooms}, "
                    f"maze_global_room_ratio: {self.maze.global_room_ratio}, "
                    f"maze_corridor_algorithm: {self.maze.corridor_algorithm}, "
                    f"maze_corridor_grid_connect_option: {self.maze.corridor_grid_connect_option}, "
                    f"maze_corridor_post_process_option: {self.maze.corridor_post_process_option}, "
                    f"maze_corridor_sort_access_points_option: {self.maze.corridor_sort_access_points_option}"
                )
                print(init_message)
                first_info_printed = False

            if done:
                first_info_printed = True
                self.reset()


if __name__ == "__main__":
    env = Labyrinth(20, 20, maze_corridor_algorithm="astar")
    env.human_play(print_info=True, animate=True)



from abc import ABC, abstractmethod, abstractclassmethod
from typing import List, Tuple, Union, Optional
import math
import random
import numpy as np

from ..constants import WALL, PATH


class RoomFactory:
    def __init__(
        self,
        access_points_nr: Optional[int] = None,
        access_points_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        ratio: Optional[Union[int, float]] = None,
        ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        seed: Optional[int] = None,
    ) -> None:
        """Construct a RoomFactory object for generating room layouts.

        Arguments `access_points_nr`, `ratio` are used to fix specific values,
        otherwise the values are drawn from the minimum and maximum distributions.

        Argument `room_types` is used to determine what types of rooms are to be added.
        If `None`, the random selection considers all the implemented room types.

        Args:
            access_points_nr (int, optional): The number of access points in the room. Defaults to None.
            access_points_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added. Defaults to None.
            ratio (float, optional): The room ratio. Defaults to None.
            ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """
        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)
        self.room_seed = self.seed

        self.py_random = random.Random(self.seed)
        self.np_random = np.random.RandomState(self.seed)

        self.access_points_nr = access_points_nr
        self.access_points_range = access_points_range

        all_available_room_types = [
            "rectangular",
        ]
        self.room_types = room_types
        if self.room_types is None:
            self.room_types = all_available_room_types

        self.ratio = ratio
        self.ratio_range = ratio_range

    def create_room(
        self,
        rows=None,
        cols=None,
        desired_area=None,
    ):
        if desired_area:
            if desired_area <= 3:
                raise ValueError(
                    f"Attempted to create a room with area {desired_area} but desired area must be greater than 3."
                )

            # Setup ratio
            if self.ratio:
                ratio = self.ratio
            else:
                ratio = self.py_random.uniform(*self.ratio_range)

            rows, cols = self._estimate_dimensions_from_area(desired_area, ratio)

        # Setup number of access points
        if self.access_points_nr:
            nr_access_points = self.access_points_nr
        else:
            nr_access_points = self.py_random.randint(
                self.access_points_range[0], self.access_points_range[1]
            )

        # increment room seed to seed each room differently
        self.room_seed += 1

        room_type = self.py_random.choice(self.room_types)

        if room_type == "rectangular":
            return self.create_rectangular_room(
                rows=rows, cols=cols, nr_access_points=nr_access_points
            )

    def _estimate_dimensions_from_area(self, desired_area, ratio=1):
        """Estimate room dimensions based on desired area and given ratio."""

        cols = int(math.sqrt(desired_area / ratio))
        rows = int(ratio * cols)

        return rows, cols

    def create_rectangular_room(self, rows=None, cols=None, nr_access_points=1):
        """Create a rectangular room using given rows, columns or desired area."""
        return RectangularRoom(
            rows=rows, cols=cols, nr_access_points=nr_access_points, seed=self.room_seed
        )


class Room(ABC):
    def __init__(
        self,
        rows: Optional[int] = None,
        cols: Optional[int] = None,
        nr_access_points: int = 1,
        min_rows: int = 3,
        min_cols: int = 3,
        seed: Optional[int] = None,
        top_left_coord: Tuple[int, int] = (0, 0),
    ) -> None:
        """
        Base Room class.

        Args:
            rows (int, optional): The number of rows in the room. Defaults to None.
            cols (int, optional): The number of columns in the room. Defaults to None.
            nr_access_points (int, optional): The number of access points for the room. Defaults to 1.
            min_rows (int, optional): Minimum rows for the room. Defaults to 3.
            min_cols (int, optional): Minimum columns for the room. Defaults to 3.
            seed (int, optional): Seed value for random operations. Defaults to None.
            top_left_coord (tuple, optional): The top left coordinate of the room. Defaults to (0, 0).
        """

        self.rows = max(rows, min_rows)
        self.cols = max(cols, min_cols)

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)

        self.np_random = np.random.RandomState(self.seed)

        self.top_left_coord = top_left_coord  
        self.bottom_right_coord = (self.rows, self.cols)
        self.nr_access_points = nr_access_points
        self.access_points = set()
        self.grid = np.ones((self.rows, self.cols), dtype=int) * WALL

    @property
    def area(self):
        return self.rows * self.cols

    @property
    def global_position(self):
        return self.top_left_coord

    @global_position.setter
    def global_position(self, position):
        self.top_left_coord = position

    @property
    def shape(self):
        return self.rows, self.cols

    @abstractmethod
    def generate_room_layout(self):
        pass

    @abstractmethod
    def get_perimeter_cells(self, padding=0):
        """Get a list of the coordinates of the perimeter cells."""
        pass

    def set_access_points(self):
        """Define default access points for the room, based on its shape.

        Access points should be represented as a set of (x, y) tuples.
        """
        perimeter_cells = self.get_perimeter_cells()
        chosen_indices = self.np_random.choice(
            len(perimeter_cells), self.nr_access_points, replace=False
        )
        self.access_points = {perimeter_cells[i] for i in chosen_indices}

    @abstractmethod
    def generate_inner_area_mask(self, total_rows, total_cols):
        pass


class RectangularRoom(Room):
    def __init__(
        self,
        rows: int = 5,
        cols: int = 5,
        nr_access_points: int = 1,
        min_rows: int = 3,
        min_cols: int = 3,
        seed: Optional[int] = None,
        **kwargs
    ) -> None:
        super().__init__(
            rows=rows,
            cols=cols,
            nr_access_points=nr_access_points,
            min_rows=min_rows,
            min_cols=min_cols,
            seed=seed,
            **kwargs
        )
        """Construct a RectangularRoom object representing a rectangular room.

        Args:
            rows (int, optional): The number of rows in the room. Defaults to 5.
            cols (int, optional): The number of columns in the room. Defaults to 5.
            nr_access_points (int, optional): The number of access points in the room. Defaults to 1.
            min_rows (int, optional): The minimum number of rows for the room. Defaults to 3.
            min_cols (int, optional): The minimum number of columns for the room. Defaults to 3.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """

        self.generate_room_layout()
        self.set_access_points()

    def generate_room_layout(self):
        self.grid[:] = PATH

    def get_perimeter_cells(self, padding=0):
        # Top perimeter with padding
        top = [(-padding, j) for j in range(-padding, self.cols + padding)]

        # Bottom perimeter with padding
        bottom = [
            (self.rows - 1 + padding, j) for j in range(-padding, self.cols + padding)
        ]

        # Left perimeter with padding (excluding the corners)
        left = [(i, -padding) for i in range(1 - padding, self.rows - 1 + padding)]

        # Right perimeter with padding (excluding the corners)
        right = [
            (i, self.cols - 1 + padding)
            for i in range(1 - padding, self.rows - 1 + padding)
        ]

        perimeter = top + bottom + left + right
        return perimeter

    def generate_inner_area_mask(self):
        return np.ones((self.rows, self.cols), dtype=bool)


class CircularRoom(Room):
    # Implement similar methods for CircularRoom, approximating a circle shape on a grid.
    pass


class DonutRoom(Room):
    # Implement similar methods for DonutRoom, approximating a donut shape on a grid.
    pass


class LShapedRoom(Room):
    # Implement similar methods for LShapedRoom, approximating an L shape on a grid.
    pass


class TShapedRoom(Room):
    pass


class TriangleRoom(Room):
    # Implement similar methods for LShapedRoom, approximating an L shape on a grid.
    pass



import numpy as np
import random
from collections import deque
from typing import List, Tuple, Union, Optional
from enum import Enum

from .room import RoomFactory
from .corridor import CorridorBuilder
from ..constants import WALL, PATH, START, TARGET, CorridorMoveStatus

from ...common.grid_functions import on_line


class MazeFactory:
    def __init__(
        self,
        rows: int,
        cols: int,
        nr_desired_rooms: Optional[int] = None,
        nr_desired_rooms_range: Tuple[int, int] = (1, 8),
        global_room_ratio: Optional[float] = None,
        global_room_ratio_range: Tuple[float, float] = (0.1, 0.8),
        access_points_per_room: Optional[int] = None,
        access_points_per_room_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        room_ratio: Optional[Union[int, float]] = None,
        room_ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        corridor_algorithm: Optional[str] = "random",
        corridor_grid_connect_option: Optional[Union[bool, str]] = "random",
        corridor_post_process_option: bool = True,
        corridor_sort_access_points_option: Optional[Union[bool, str]] = "random",
        seed: Optional[int] = None,
    ):
        """
        Maze builder to randomize the generated maze.

        Arguments num_rooms, global_room_ratio, access_points_per_room
        are used to fix specific values, otherwise the values are drawn from the minimum and maximum distibutions.

        room_types is used to determine what types of rooms are to be added in the maze, if None, the random seletion
        considers all the implemented room types.

        Args:
            rows (int): The number of rows in the maze.
            cols (int): The number of columns in the maze.
            nr_desired_rooms (int, optional): The desired number of rooms in the maze. Defaults to None.
            nr_desired_rooms_range (tuple, optional): The range of desired number of rooms. Defaults to (1, 8).
            global_room_ratio (float, optional): The global room ratio in the maze. Defaults to None.
            global_room_ratio_range (tuple, optional): The range of global room ratio. Defaults to (0.1, 0.8).
            access_points_per_room (int, optional): The number of access points in each room. Defaults to None.
            access_points_per_room_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added in the maze. Defaults to None.
            room_ratio (float, optional): The room ratio. Defaults to None.
            room_ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            corridor_algorithm (str, optional): The algorithm to use for generating the maze. Defaults to 'random'. Can be one of: 'prim', 'astar', 'gbfs' or 'random'. if 'random', randomly selects algorithm.
            corridor_grid_connect_option (Union[bool, str], optional): Option to decide the nature of corridor paths connectivity.
                - If `True`, corridors will be grid-connected.
                - If `False`, corridors will not be grid-connected.
                - If `"random"`, the choice will be made randomly.
                Defaults to False.
            corridor_post_process_option (bool, optional): Whether to post-process the maze. Defaults to True.
            corridor_sort_access_points_option (Union[bool, str], str, optional): Whether to sort the access points. Defaults to "random". If "random" will randomly select between True and False.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """

        self.rows = rows
        self.cols = cols

        self.nr_desired_rooms = nr_desired_rooms
        self.nr_desired_rooms_range = nr_desired_rooms_range

        self.global_room_ratio = global_room_ratio
        self.global_room_ratio_range = global_room_ratio_range

        self.access_points_per_room = access_points_per_room
        self.access_points_per_room_range = access_points_per_room_range

        self.room_types = room_types

        self.room_ratio = room_ratio
        self.room_ratio_range = room_ratio_range

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)
        self.py_random = random.Random(self.seed)
        self.np_random = np.random.RandomState(self.seed)

        self.corridor_algorithm = corridor_algorithm
        self._check_corridor_algorithm_option()

        self.corridor_grid_connect_option = corridor_grid_connect_option
        self._check_corridor_grid_connect_option()

        self.corridor_post_process_option = corridor_post_process_option

        self.corridor_sort_access_points_option = corridor_sort_access_points_option
        self._check_corridor_sort_access_points_option()

    def create_maze(self):
        # Decide the number of rooms
        if self.nr_desired_rooms:
            num_rooms = self.nr_desired_rooms
        else:
            num_rooms = self.py_random.randint(
                self.nr_desired_rooms_range[0], self.nr_desired_rooms_range[1]
            )

        # Decide the global room ratio
        if self.global_room_ratio:
            if self.global_room_ratio > 1:
                raise ValueError(
                    "Global room ratio must be less than 1, but got {self.global_room_ratio}."
                )
            global_room_ratio = self.global_room_ratio
        else:
            if (
                self.global_room_ratio_range[0] < 0
                or self.global_room_ratio_range[1] > 1
            ):
                raise ValueError(
                    f"Global room ratio range must be between 0 and 1, but got {self.global_room_ratio_range}."
                )
            global_room_ratio = self.py_random.uniform(
                self.global_room_ratio_range[0], self.global_room_ratio_range[1]
            )

        # Randomize or fix other parameters
        self._check_corridor_algorithm_option()
        if self.corridor_algorithm == "random":
            corridor_algorithm = self.py_random.choice(["prim", "astar"])
        else:
            corridor_algorithm = self.corridor_algorithm

        self._check_corridor_grid_connect_option()
        if self.corridor_grid_connect_option == "random":
            corridor_grid_connect_option = self.py_random.choice([True, False])
        else:
            corridor_grid_connect_option = self.corridor_grid_connect_option

        self._check_corridor_sort_access_points_option()
        if self.corridor_sort_access_points_option == "random":
            corridor_sort_access_points_option = self.py_random.choice([True, False])
        else:
            corridor_sort_access_points_option = self.corridor_sort_access_points_option

        maze_seed = self.seed + 1
        maze = Maze(
            rows=self.rows,
            cols=self.cols,
            nr_desired_rooms=num_rooms,
            global_room_ratio=global_room_ratio,
            access_points_per_room=self.access_points_per_room,
            access_points_per_room_range=self.access_points_per_room_range,
            room_types=self.room_types,
            room_ratio=self.room_ratio,
            room_ratio_range=self.room_ratio_range,
            corridor_algorithm=corridor_algorithm,
            corridor_grid_connect_option=corridor_grid_connect_option,
            corridor_post_process_option=self.corridor_post_process_option,
            corridor_sort_access_points_option=corridor_sort_access_points_option,
            seed=maze_seed,
        )
        return maze

    def _check_corridor_grid_connect_option(self):
        if self.corridor_grid_connect_option not in [True, False, "random"]:
            raise ValueError(
                f"Invalid value for corridor_grid_connect_option. Got {self.corridor_grid_connect_option}, expected [True, False, 'random']."
            )

    def _check_corridor_algorithm_option(self):
        if self.corridor_algorithm not in ["prim", "astar", "gbfs", "random"]:
            raise ValueError(
                f"Invalid value for corridor_algorithm. Got {self.corridor_algorithm}, expected ['prim', 'astar', 'random']."
            )

    def _check_corridor_sort_access_points_option(self):
        if self.corridor_sort_access_points_option not in [True, False, "random"]:
            raise ValueError(
                f"Invalid value for corridor_sort_access_points_option. Got {self.corridor_sort_access_points_option}, expected [True, False, 'random']."
            )


class Maze:
    def __init__(
        self,
        rows: int,
        cols: int,
        nr_desired_rooms: int = 3,
        global_room_ratio: float = 0.5,
        access_points_per_room: Optional[int] = None,
        access_points_per_room_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        room_ratio: Optional[Union[int, float]] = None,
        room_ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        corridor_algorithm: Optional[str] = "prim",
        corridor_grid_connect_option: Optional[bool] = False,
        corridor_post_process_option: Optional[bool] = True,
        corridor_sort_access_points_option: Optional[bool] = False,
        seed: Optional[int] = None,
    ) -> None:
        """Construct a Maze object representing a maze layout.

        Args:
            rows (int): The number of rows in the maze.
            cols (int): The number of columns in the maze.
            nr_desired_rooms (int, optional): The desired number of rooms in the maze. Defaults to 3.
            global_room_ratio (float, optional): The global room ratio in the maze. Defaults to 0.5.
            access_points_per_room (int, optional): The number of access points in each room. Defaults to None.
            access_points_per_room_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added in the maze. Defaults to None.
            room_ratio (float, optional): The room ratio. Defaults to None.
            room_ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            algorithm (str, optional): The algorithm to use for generating the maze. Defaults to 'prim'. Can be one of: 'prim', 'astar'.
            corridor_grid_connect_option (bool, optional): Option to decide the nature of corridor paths connectivity.
                - If `True`, corridors will be grid-connected.
                - If `False`, corridors will not be grid-connected.
                Defaults to False. Only affects 'prim' algorithm.
            corridor_grid_connect_option (bool, optional): Whether to post-process the maze. Defaults to True. Only affects 'prim' algorithm.
            corridor_sort_access_points_option (bool, optional): Whether to sort the access points in corridor generation. Defaults to False.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """
        if rows < 10 or cols < 10:
            raise ValueError(
                "Maze dimensions must be at least 10x10, otherwise the maze generation logic risks breaking."
            )

        self.rows = rows
        self.cols = cols

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)
        self.py_random = random.Random(self.seed)

        self.grid = np.ones((rows, cols), dtype=int) * WALL
        self.room_grid = np.ones((rows, cols), dtype=int) * WALL
        self.corridor_grid = np.ones((rows, cols), dtype=int) * WALL

        self.rooms = []
        self.nr_desired_rooms = nr_desired_rooms

        # in some cases we won't be able to place all rooms we want so track them
        self.nr_placed_rooms = 0
        self.global_room_ratio = global_room_ratio

        self.start_position = None
        self.target_position = None

        # Make Room Factory
        room_factory_seed = self.seed + 1
        self.room_factory = RoomFactory(
            access_points_nr=access_points_per_room,
            access_points_range=access_points_per_room_range,
            room_types=room_types,
            ratio=room_ratio,
            ratio_range=room_ratio_range,
            seed=room_factory_seed,
        )

        # Corridor generationg settings
        self.corridor_algorithm = corridor_algorithm
        self.corridor_grid_connect_option = corridor_grid_connect_option
        self.corridor_post_process_option = corridor_post_process_option
        self.corridor_sort_access_points_option = corridor_sort_access_points_option

        # Make Corridor Builder
        self.corridor_builder = CorridorBuilder(self)

        self._build_maze()

    def _build_maze(self):
        self.make_rooms()
        self.grid = np.where(self.room_grid == PATH, PATH, self.grid)

        self.make_corridors()
        self.grid = np.where(self.corridor_grid == PATH, PATH, self.grid)

    def make_corridors(self):
        self.place_start_end_positions()

        if self.corridor_algorithm == "prim":
            self.corridor_grid = self.corridor_builder.generate_corridor_prim()
            self.corridor_builder.connect_rooms_to_paths()
            if self.corridor_post_process_option:
                self.corridor_builder.post_process_maze()

        elif self.corridor_algorithm == "astar":
            self.corridor_grid = self.corridor_builder.generate_corridor_a_star()

        elif self.corridor_algorithm == "gbfs":
            self.corridor_grid = self.corridor_builder.generate_corridor_gbfs()

    ##### Room generation and placement #####
    def levy_flight_place(self, room):
        """Seek a random position for a room in such a way
        that it does not overlap with other rooms."""
        max_attempts = 100
        attempt = 0

        margin_padding = 2

        # define range from 2 to len - (minsize + 2) because we don't want to place rooms too close to edge of maze
        row_padding = room.rows + margin_padding
        col_padding = room.cols + margin_padding

        bottom_right_possible_row_coord = self.grid.shape[0] - row_padding
        bottom_right_possible_col_coord = self.grid.shape[1] - col_padding

        if (
            bottom_right_possible_row_coord < margin_padding
            or bottom_right_possible_col_coord < margin_padding
        ):
            # This proposed room is too big
            return False

        #### Attempt to find a viable position ####
        position = (
            self.py_random.randint(margin_padding, bottom_right_possible_row_coord),
            self.py_random.randint(margin_padding, bottom_right_possible_col_coord),
        )
        directions = [(0, 1), (0, -1), (1, 0), (-1, 0)]

        while attempt < max_attempts:
            if self.is_valid_room_position(room, position[0], position[1]):
                self.rooms.append(room)
                room.global_position = position
                self.materialize_room(room, position[0], position[1])
                return True
            else:
                step_size = self.levy_step_size()
                direction = self.py_random.choice(directions)
                position = (
                    position[0] + direction[0] * step_size,
                    position[1] + direction[1] * step_size,
                )
                attempt += 1

        return False

    def make_rooms(self):
        """Randomly generate rooms and attempt to place them in the maze grid."""
        total_room_path_area_covered = 0
        target_path_area = self.global_room_ratio * self.grid.size
        start_avg_room_path_area = target_path_area / self.nr_desired_rooms
        desired_room_path_area = start_avg_room_path_area

        # Because of the desired_room_path_area = max(9, desired_room_path_area),
        # we need to also check for a maximum attempt counter
        attempts = 0
        max_attempts = 100

        room_sizes = []
        while (
            self.nr_placed_rooms < self.nr_desired_rooms
            and total_room_path_area_covered < target_path_area
            and attempts < max_attempts
        ):
            # In some cases when the global ratio is small and we want many rooms,
            # the start_avg is small from the very start so we need to enforce minimum area
            desired_room_path_area = max(9, desired_room_path_area)

            room = self.room_factory.create_room(desired_area=desired_room_path_area)
            room_sizes.append((room.rows, room.cols))
            if self.levy_flight_place(room):
                total_room_path_area_covered += room.rows * room.cols
                self.nr_placed_rooms += 1

            # Each time a room is added, or a fail happens, decrease the desired area of the next room trials
            desired_room_path_area = int(0.9 * desired_room_path_area)
            attempts += 1

    def is_valid_room_position(self, room, start_row, start_col):
        end_row, end_col = start_row + room.rows, start_col + room.cols

        # Extract sub-grid and check boundaries
        if (
            start_row < 1
            or start_col < 1
            or end_row > self.room_grid.shape[0] - 1
            or end_col > self.room_grid.shape[1] - 1
        ):
            return False

        # Basically add a padding of 1
        sub_grid = self.room_grid[
            start_row - 1 : end_row + 1, start_col - 1 : end_col + 1
        ]

        # Check if placing the room would overlap with another room or not maintain separation
        return np.sum(sub_grid == PATH) == 0

    def materialize_room(self, room, start_row, start_col):
        self.room_grid[
            start_row : start_row + room.rows, start_col : start_col + room.cols
        ] = room.grid

    def levy_step_size(self):
        r = self.py_random.random()  # r is between 0 and 1
        return int(
            1 / ((r + 0.001) ** 0.5)
        )  # This will give us a step length L according to inverse square distribution.

    ##### Corridor generation #####

    def choose_start_position(self):
        """Choose a random position on the perimeter of the maze grid."""
        side = self.py_random.choice(["top", "bottom", "left", "right"])

        if side == "top":
            return (0, self.py_random.randint(0, self.grid.shape[1] - 1))
        elif side == "bottom":
            return (
                self.grid.shape[0] - 1,
                self.py_random.randint(0, self.grid.shape[1] - 1),
            )
        elif side == "left":
            return (self.py_random.randint(0, self.grid.shape[0] - 1), 0)
        else:  # 'right'
            return (
                self.py_random.randint(0, self.grid.shape[0] - 1),
                self.grid.shape[1] - 1,
            )

    def choose_target_position(self):
        """Choose a random target position in one of the generated rooms."""

        # Shuffle the rooms to ensure random selection without replacement
        shuffled_rooms = self.py_random.sample(self.rooms, len(self.rooms))

        for random_room in shuffled_rooms:
            room_top_left_row, room_top_left_col = random_room.global_position

            # Create a list of all possible cell positions in the room
            all_positions = [
                (row, col)
                for row in range(random_room.rows)
                for col in range(random_room.cols)
            ]
            # Shuffle these positions to randomly select without replacement
            self.py_random.shuffle(all_positions)

            for row, col in all_positions:
                # Target must be on an empty room tile and not in the perimeter
                if (random_room.grid[row, col] == PATH) and (
                    (row, col) not in random_room.get_perimeter_cells()
                ):
                    return (room_top_left_row + row, room_top_left_col + col)

        # If the function hasn't returned by this point, then no valid target position was found in any room
        raise ValueError(f"Could not find a valid target position in any room.")

    def place_start_end_positions(self):
        self.start_position = self.choose_start_position()
        self.target_position = self.choose_target_position()

    def is_inside_any_room(self, pos, exception=None):
        """Check if a position is inside any of the rooms in the maze."""

        for room in self.rooms:
            # Check if the position is within the global boundaries of the room
            if (
                room.global_position[0] <= pos[0] < room.global_position[0] + room.rows
                and room.global_position[1]
                <= pos[1]
                < room.global_position[1] + room.cols
            ):
                # If the position is the exception, consider it not inside the room
                if exception and pos == exception:
                    continue
                # Otherwise, it's inside this room
                return True

        # If we reached here, the position isn't inside any room
        return False

    def generate_global_room_mask(self):
        global_room_mask = np.zeros((self.rows, self.cols), dtype=bool)

        for room in self.rooms:
            local_mask = room.generate_inner_area_mask()
            global_position = room.global_position

            for i in range(local_mask.shape[0]):
                for j in range(local_mask.shape[1]):
                    global_room_mask[
                        global_position[0] + i, global_position[1] + j
                    ] |= local_mask[i, j]

        return global_room_mask

    #### Validate maze ####
    def is_valid_maze(self):
        """Validate if all access points are reachable from the starting point.
        Sanity check function, was not included in the constructor but can be used after generation.
        """
        start_point = self.start_position
        visited = set([start_point])
        queue = deque([start_point])

        while queue:
            current_point = queue.popleft()
            for neighbor in self.get_neighbors(*current_point):
                if neighbor not in visited:
                    visited.add(neighbor)
                    queue.append(neighbor)

        # Check if all PATH cells have been visited
        all_path_cells = {
            (x, y)
            for x in range(self.rows)
            for y in range(self.cols)
            if self.grid[x][y] == PATH
        }

        return visited == all_path_cells

    def get_neighbors(self, x, y):
        """Get neighboring PATH cells of a given cell (x, y)."""
        neighbors = []
        for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
            nx, ny = x + dx, y + dy
            if (
                0 <= nx < self.rows
                and 0 <= ny < self.cols
                and self.grid[nx][ny] == PATH
            ):
                neighbors.append((nx, ny))
        return neighbors



from enum import IntEnum

class Action(IntEnum):
    UP = 0
    RIGHT = 1
    DOWN = 2
    LEFT = 3

class CorridorMoveStatus(IntEnum):
    VALID_MOVE = 1
    MAZE_BOUNDARY = 2
    ROOM_BOUNDARY = 3
    INVALID = 4

### Constants for states in the maze
WALL = 0
PATH = 1 
TARGET = 2 
START = 3
PLAYER = 4

